import yaml
import tilekiln.layer
import json


class Config:
    def __init__(self, yaml_string, fs):
        '''Create a config from a yaml string
           Creates a config from the yaml string. Any SQL files referenced
           must be in the filesystem.
        '''
        config = yaml.safe_load(yaml_string)

        # There's a lot of metadata properties, so pull that table out
        md = config["metadata"]
        self.id = md["id"]
        self.name = md.get("name")
        self.description = md.get("description")
        self.attribution = md.get("attribution")
        self.version = md.get("version")
        self.bounds = md.get("bounds")
        self.center = md.get("center")

        self.layers = []
        # Make vector_layers optional. TODO: Reconsider
        for id, l in config.get("vector_layers", {}).items():
            self.layers.append(tilekiln.layer.Layer(id, l, fs))

    @property
    def minzoom(self):
        return min([layer.minzoom for layer in self.layers])

    @property
    def maxzoom(self):
        return max([layer.maxzoom for layer in self.layers])

    def tilejson(self, url):
        '''Returns a tilejson as a string for the config
        '''

        # See https://github.com/mapbox/tilejson-spec/tree/master/2.2.0

        # Work by assembling a dictionary then turning it into json

        # Required properties
        tj = {"tilejson": "2.2.0",
              "format": "pbf",
              "scheme": "xyz",
              "tiles": [url.replace("{id}", self.id)]}

        if self.name is not None:
            tj["name"] = self.name
        if self.description is not None:
            tj["description"] = self.description
        if self.attribution is not None:
            tj["attribution"] = self.attribution
        if self.version is not None:
            tj["version"] = self.version
        if self.bounds is not None:
            tj["bounds"] = self.bounds
        if self.center is not None:
            tj["center"] = self.center

        if len(self.layers) > 0:
            tj["minzoom"] = min([layer.minzoom for layer in self.layers])
            tj["maxzoom"] = max([layer.maxzoom for layer in self.layers])

            tj["vector_layers"] = []
            for layer in self.layers:
                tj_layer = {"id": layer.id}
                if layer.description is not None:
                    tj_layer["description"] = layer.description
                tj_layer["minzoom"] = layer.minzoom
                tj_layer["maxzoom"] = layer.maxzoom

                tj_layer["fields"] = layer.fields

                if layer.geometry_type == set(["polygon"]):
                    tj_layer["geometry_type"] = "polygon"
                elif layer.geometry_type == set(["point"]):
                    tj_layer["geometry_type"] = "point"
                elif layer.geometry_type == set(["line"]):
                    tj_layer["geometry_type"] = "line"
                else:
                    tj_layer["geometry_type"] = "unknown"

                tj["vector_layers"].append(tj_layer)

        return json.dumps(tj, sort_keys=True, indent=4)



import click
import fs.osfs
import tilekiln.config
import tilekiln.kiln
import os


@click.command()
@click.argument('config', type=click.Path(exists=True))
@click.argument('url', default='http://localhost/{id}/{z}/{x}/{y}.mvt')
def cli(config, url):
    # Get the directory the config is in
    # TODO: Reduce duplication between this and tilejson-generate
    full_path = os.path.join(os.getcwd(), config)
    root_path = os.path.dirname(full_path)
    config_path = os.path.relpath(full_path, root_path)

    filesystem = fs.osfs.OSFS(root_path)

    config = tilekiln.config.Config(filesystem.open(config_path).read(),
                                    filesystem)
    print(config.tilejson(url))



import jinja2 as j2


# Invariants of web mercator
HALF_WORLD = 20037508.34


class Definition:
    ''' Definition of a layer in a particular zoom range.

    This class does the work of producing the SQL query for a particular tile.
    '''
    def __init__(self, id, raw_sql, minzoom, maxzoom, extent):
        self.id = id
        self.raw_sql = raw_sql
        self.minzoom = minzoom
        self.maxzoom = maxzoom

        self.extent = extent or 4096

    def __eq__(self, other):
        return (self.id == other.id and self.raw_sql == other.raw_sql and
                self.minzoom == other.minzoom and
                self.maxzoom == other.maxzoom and
                self.extent == other.extent)

    def __repr__(self):
        return ('Definition({}, "{}", {}, {}, {})'
                .format(self.id, self.raw_sql, self.minzoom, self.maxzoom,
                        self.extent))

    def render_sql(self, tile):
        '''Generates the SQL for a layer
        '''
        zoom, x, y = tile
        assert(zoom >= self.minzoom)
        assert(zoom <= self.maxzoom)
        assert(x >= 0)
        assert(x < 2**zoom)
        assert(y >= 0)
        assert(y < 2**zoom)

        # See See https://postgis.net/docs/ST_AsMVT.html for SQL source
        # TODO: Add parameters for all ST_AsMVT and ST_AsMVTGeom options
        sql = j2.Template(self.raw_sql)

        inner = sql.render(zoom=zoom, x=x, y=y,
                           tile_length=tile_length(zoom),
                           tile_area=tile_area(zoom),
                           coordinate_length=coordinate_length(zoom,
                                                               self.extent),
                           coordinate_area=coordinate_area(zoom,
                                                           self.extent),
                           extent=self.extent,
                           bbox=bbox(zoom, x, y))

        return wrap_sql(inner, self.id, self.extent)


# Utility functions
def wrap_sql(sql, layer_id, extent):
    return ('''WITH mvtgeom AS\n(\n''' + sql + '''\n)\n''' +
            '''SELECT ST_AsMVT(mvtgeom.*, '{}', '''.format(layer_id) +
            '''{}, 'way', NULL)\n'''.format(extent) +
            '''FROM mvtgeom;''')


def bbox(zoom, x, y):
    ''' Returns an array of [minx, miny, max, maxy]
        There is a flip of y axis converting between xyz and EPSG:3857
    '''
    ll = zxy_to_projected(zoom, x, y)
    ur = zxy_to_projected(zoom, x+1, y+1)
    return ('ST_MakeEnvelope({}, {}, {}, {}, 3857)'
            .format(ll[0], ur[1], ur[0], ll[1]))


def tile_length(zoom):
    # -1 for half vs full world
    return HALF_WORLD/(2**(zoom-1))


def coordinate_length(zoom, extent):
    return tile_length(zoom)/extent


def tile_area(zoom):
    return tile_length(zoom)**2


def coordinate_area(zoom, extent):
    return coordinate_length(zoom, extent)**2


# ref https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
#     https://github.com/mapbox/postgis-vt-util/blob/master/src/TileBBox.sql
#     and others.
def zxy_to_projected(zoom, x, y):
    return [HALF_WORLD*(2*x/2**zoom - 1),
            HALF_WORLD*(1-2*y/2**zoom)]



from tilekiln.definition import Definition


class Layer:
    '''A layer in the tileset

    A Layer contains all the information needed to say what the definitions
    are for every zoom
    '''
    def __init__(self, id, layer_config, fs):
        self.id = id
        self.description = layer_config.get("description")
        self.fields = layer_config.get("fields")

        self.definitions = []

        self.geometry_type = set(layer_config.get("geometry_type", []))

        if "sql" in layer_config:
            for d in layer_config["sql"]:
                self.definitions.append(Definition(self.id,
                                        fs.readtext(d["file"]),
                                        d["minzoom"], d["maxzoom"],
                                        d.get("extent")))

            self.minzoom = min([d.minzoom for d in self.definitions])
            self.maxzoom = max([d.maxzoom for d in self.definitions])

    def __eq__(self, other):
        return (self.id == other.id and self.description == other.description
                and self.fields == other.fields
                and self.definitions == other.definitions
                and self.geometry_type == other.geometry_type)

    def definition_for_zoom(self, zoom):
        ''' Get the right definition for a given zoom
        '''
        for d in self.definitions:
            if zoom >= d.minzoom and zoom <= d.maxzoom:
                return d

    def render_tile(self, tile, db):
        d = self.definition_for_zoom(tile[0])
        if d is None:
            return None

        return db.generate_tilelayer(d, tile)



import multiprocessing.pool
import tilekiln.generator


class Kiln:
    ''' Class that makes tiles
    '''
    def __init__(self, config, dbconn, storage_location):
        '''Create the kiln for tiles

        config - Config object
        dbinfo - database connection params
        storage_location - storage location
        connections - number of connections
        '''

        self.config = config
        self.dbconn = dbconn
        self.storage_location = storage_location

    def generate_tiles(self, tiles, connections, chunk_size):
        chunks = [tiles[i * chunk_size:(i+1) * chunk_size]
                  for i in range((len(tiles) + chunk_size - 1) // chunk_size)]

        gen = tilekiln.generator.Generator(self.config.id, self.config.layers,
                                           self.dbconn, self.storage_location)

        with multiprocessing.pool.Pool(connections) as p:
            p.map(gen.write_tiles, chunks)



import click
import fs.osfs
import tilekiln.config
import tilekiln.kiln
import os
import multiprocessing
import sys


@click.group()
def cli():
    pass


@cli.command()
@click.argument('config', type=click.Path(exists=True))
@click.argument('storage')
@click.option('-d', '--dbname')
@click.option('-h', '--host')
@click.option('-p', '--port')
@click.option('-U', '--username')
@click.option('-c', '--connections', type=click.INT,
              default=multiprocessing.cpu_count())
@click.option('-s', '--chunk-size', type=click.INT)
@click.option('-z', '--min-zoom', type=click.INT)
@click.option('-Z', '--max-zoom', type=click.INT)
def area(config, storage, dbname, host, port, username, connections,
         chunk_size, min_zoom, max_zoom):
    '''Generates tiles for an area'''
    # Get the directory the config is in
    full_path = os.path.join(os.getcwd(), config)
    root_path = os.path.dirname(full_path)
    config_path = os.path.relpath(full_path, root_path)

    filesystem = fs.osfs.OSFS(root_path)

    config = tilekiln.config.Config(filesystem.open(config_path).read(),
                                    filesystem)
    dbinfo = {"dbname": dbname, "host": host, "port": port,
              "username": username}

    min_zoom = min_zoom or config.minzoom
    max_zoom = max_zoom or config.maxzoom

    tiles = [(z, x, y) for z in range(min_zoom, max_zoom + 1)
             for x in range(2**z) for y in range(2**z)]

    # Apply some heuristics to guess a chunk size
    if chunk_size is None:
        chunk_size = int(min(max(len(tiles)/(2*connections), 10), 50000))

    kiln = tilekiln.kiln.Kiln(config, dbinfo, storage)
    kiln.generate_tiles(tiles, connections, chunk_size)


@cli.command()
@click.argument('config', type=click.Path(exists=True))
@click.argument('storage')
@click.option('-d', '--dbname')
@click.option('-h', '--host')
@click.option('-p', '--port')
@click.option('-U', '--username')
@click.option('-c', '--connections', type=click.INT,
              default=multiprocessing.cpu_count())
@click.option('-s', '--chunk-size', type=click.INT)
def tiles(config, storage, dbname, host, port, username, connections,
          chunk_size):
    '''Generate a list of tiles

    Feed the tiles in on stdin and send an EOF to start generation
    '''
    # Get the directory the config is in
    full_path = os.path.join(os.getcwd(), config)
    root_path = os.path.dirname(full_path)
    config_path = os.path.relpath(full_path, root_path)

    filesystem = fs.osfs.OSFS(root_path)

    config = tilekiln.config.Config(filesystem.open(config_path).read(),
                                    filesystem)
    dbinfo = {"dbname": dbname, "host": host, "port": port,
              "username": username}

    tiles = [tuple(int(coord) for coord in line.strip().split('/', 3))
             for line in sys.stdin.readlines() if line.strip() != '']

    if chunk_size is None:
        chunk_size = int(min(max(len(tiles)/(2*connections), 10), 50000))

    kiln = tilekiln.kiln.Kiln(config, dbinfo, storage)
    kiln.generate_tiles(tiles, connections, chunk_size)



import psycopg2
import fs
import tilekiln.database


class Generator:
    def __init__(self, id, layers, dbconn, storage_location):
        '''Sets up the generator
           to generate using dbconn as specified in config and write to
           storage
        '''
        self.id = id
        self.layers = layers
        self.dbconn = dbconn
        self.storage_location = storage_location

    def write_tiles(self, tiles):
        ''' Connect to DB & storage and generate & write tiles
        '''
        db = tilekiln.database.Database(psycopg2.connect(**self.dbconn))
        filesystem = fs.open_fs(self.storage_location, create=True)

        for tile in tiles:
            tiledata = b''

            # TODO: Go through layers
            for layer in self.layers:
                layerdata = layer.render_tile(tile, db)
                if layerdata is not None:
                    tiledata += layerdata

            zoom = tile[0]
            x = tile[1]
            y = tile[2]
            output_dir = "{id}/{z}/{x}/".format(id=self.id, z=zoom, x=x)
            filesystem.makedirs(output_dir, recreate=True)
            output_path = output_dir + "{y}.mvt".format(y=y)
            filesystem.writebytes(output_path, tiledata)



from setuptools import find_packages, setup


setup(
    name='tilekiln',
    version='0.0.5',
    author="Paul Norman",
    author_email="osm@paulnorman.ca",
    url="https://github.com/pnorman/tilekiln",
    packages=find_packages(),
    include_package_data=False,
    zip_safe=False,
    install_requires=[
        'Click',
        'fs',
        'pyyaml',
        'jinja2',
        'psycopg2'
    ],
    entry_points={
        'console_scripts': ['tilekiln-generate=tilekiln.scripts.generate:cli',
                            'tilekiln-tilejson=tilekiln.scripts.tilejson:cli']
    },
    setup_requires=[
        'pytest-runner',
        'flake8'
    ],
    tests_require=['pytest'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS"
    ],
    python_requires="~=3.6"
)



class Database:
    '''A wrapper for a psycopg2 connection

    A Database produces a tile-layer, given a definition
    '''
    def __init__(self, conn):
        self.conn = conn

    def generate_tilelayer(self, definition, tile):
        with self.conn.cursor() as curs:
            sql = definition.render_sql(tile)
            curs.execute(sql)
            result = curs.fetchone()
            mvt = result[0]

        return mvt
